@@ -11,7 +11,7 @@
 #' @param n Numbers of samples (cells in scRNA-seq), nrow(X).
 #' @param d A vector of (m1,...,mk), mi = ncol(Zi), number of columns in Zi. m1 + ... + mk = ncol(Z), number of columns in Z.
 #' @param theta0 A vector of initial values of the variance components, (s1, ...,sk, s_(k+1)), si = sigma_i^2, the variance component of the i-th type random effects. s_(k+1) = sigma^2, the variance component of model residual error.
-#' @param method The REML with Fisher scoring (FS) iterative algorithm, REML-FS.
+#' @param method Either REML or ML with Fisher scoring (FS) iterative algorithm.
 #' @param max.iter The maximal number of iterations for the iterative algorithm.
 #' @param epsilon Positive convergence tolerance. If the absolute value of the first partial derivative of log likelihood is less than epsilon, the iterations converge.
 #' @param output.cov If TRUE, output the covariance matrices for the estimated coefficients, which are needed for testing contrasts.
@@ -19,6 +19,7 @@
 #'
 #' @return A list containing the following components:
 #' @return dlogL First partial derivatives of log-likelihoods for each feature (gene).
+#' @return logLik Either log maximum likelihood (ML) or log restricted maximum likelihood (REML) for each feature (gene). The log restricted maximum likelihood may have a constant difference from other computations. The constant difference is related to the fixed design matrix, X.
 #' @return niter Nmbers of iterations for each feature (gene).
 #' @return coef A matrix of estimated coefficients (fixed effects), each column corresponds to a feature (gene) and each row one covariate.
 #' @return se A matrix of the standard errors of the estimated coefficients.
@@ -65,14 +66,15 @@
 #' str(fit)
 #'
 #' @export
-lmm <- function(XX, XY, ZX, ZY, ZZ, Ynorm, n, d = ncol(ZZ), theta0 = NULL, method = "REML-FS", max.iter = 50, epsilon = 1e-5, output.cov = TRUE, output.RE = FALSE)
+lmm <- function(XX, XY, ZX, ZY, ZZ, Ynorm, n, d = ncol(ZZ), theta0 = NULL, method = c("REML", "ML"), max.iter = 50, epsilon = 1e-5, output.cov = TRUE, output.RE = FALSE)
 {
 stopifnot(!any(is.na(XY)), !any(is.na(ZX)), !any(is.na(ZY)))
+stopifnot(sum(d) == ncol(ZZ))
+
+method <- match.arg(method)
 p <- ncol(ZX)
 k <- length(d)
 
-stopifnot(sum(d) == ncol(ZZ))
-
 XXinv <- try(chol2inv(chol(XX)), silent = TRUE)
 if (inherits(XXinv, "try-error")) {
 	stop("XX is not positive-definite or X is not full column rank.")
@@ -83,8 +85,17 @@ zrz <- ZZ - ZX%*%(XXinv%*%t(ZX))
 zry <- ZY - ZX%*%(XXinv%*%XY)
 yry <- Ynorm - colSums(XY*(XXinv%*%XY))
 
+if (method == "REML"){
+	pres <- p
+	ZZres <- zrz
+} else {
+	pres <- 0
+	ZZres <- ZZ
+}
+
 niter <- NULL
 dlogL <- NULL
+loglike <- NULL
 theta <- matrix(nrow = k + 1, ncol = ncol(XY), dimnames = list(paste0("var", c(1:k, 0)), colnames(XY)))
 setheta <- theta
 RE <- matrix(nrow = nrow(ZY), ncol = ncol(XY), dimnames = dimnames(ZY))
@@ -98,11 +109,12 @@ for (jy in 1:ncol(ZY)) {
 		s <- c(rep(0, k), yry[jy]/(n-p))
 	} else s <- theta0
 
-vest <- varest(xxz, zrz, zryj = zry[, jy], yryj = yry[jy], n = n, d = d, s = s, max.iter = max.iter, epsilon = epsilon)
+vest <- varest(ZZres, zrz, zryj = zry[, jy], yryj = yry[jy], n = n, d = d, s = s, p = pres, max.iter = max.iter, epsilon = epsilon)
 s <- vest$s
 dl <- vest$dl
 iter <- vest$iter
 Minv <- vest$Minv
+logdet <- vest$logdetM0
 
 if (max(abs(dl)) > epsilon) {
 	warningText <- paste0("The first derivatives of log likelihood for Y", jy)
@@ -112,6 +124,9 @@ if (max(abs(dl)) > epsilon) {
 
 sr <- s[1:k]/s[k+1]
 M <- solve(sweep(ZZ, 1, STATS = rep(sr, times = d), FUN = "*") + diag(sum(d)))
+#qrM0 <- qr(M)
+#logdet <- sum(log(abs(diag(qrM0$qr))))
+
 M <- sweep(M, 2, STATS = rep(sr, times = d), FUN = "*")
 xvx <- XXinv + xxz%*%(ginv(diag(sum(d)) - M%*%(ZX%*%xxz))%*%(M%*%t(xxz)))
 xvy <- XY[, jy] - t(ZX)%*%(M%*%ZY[, jy])
@@ -125,6 +140,7 @@ theta[, jy] <- s
 setheta[, jy] <- sqrt(diag(Minv))
 beta[, jy] <- b
 dlogL <- cbind(dlogL, dl)
+loglike <- c(loglike, -(n-pres)*(1+log(2*pi*s[k+1]))/2 + logdet/2)
 sebeta[, jy] <- sqrt(diag(as.matrix(covbeta[,,jy])))
 }
 tval <- beta/sebeta
@@ -133,21 +149,23 @@ pval <- 2 * pt(-abs(tval), df = n-p)
 if (!output.cov) covbeta <- NULL
 if (!output.RE) RE <- NULL
 
-list(method = method, dlogL = dlogL, niter = niter, coef = beta, se = sebeta, t = tval, p = pval, cov = covbeta, df = n-p, theta = theta, se.theta = setheta, RE = RE)
+list(method = method, dlogL = dlogL, logLik = loglike, niter = niter, coef = beta, se = sebeta, t = tval, p = pval, cov = covbeta, df = n-p, theta = theta, se.theta = setheta, RE = RE)
 }
 
 
+
 #' A internal function to estimate variance components for one feature (gene).
 #'
 #' @description This function is used internally (inside lmm).
 #'
-#' @param xxz XXinv\%*\%t(ZX), where XXinv is the inverse of XX and ZX = t(Z)\%*\%X.
-#' @param zrz ZZ - ZX\%*\%(XXinv\%*\%t(ZX)), ZZ = t(Z)\%*\%Z.
+#' @param ZZres Equal to t(Z)\%*\%Z for ML, otherwise, t(Z)\%*\%Z - ZX\%*\%(XXinv\%*\%t(ZX)) for REML.
+#' @param zrz t(Z)\%*\%Z - ZX\%*\%(XXinv\%*\%t(ZX)).
 #' @param zryj zry[, j], where zry = ZY - ZX\%*\%(XXinv\%*\%XY)
 #' @param yryj yry[j], where yry = Ynorm - colSums(XY*(XXinv\%*\%XY))
 #' @param n Numbers of samples (cells in scRNA-seq).
 #' @param d A vector of (m1,...,mk), mi = ncol(Zi), number of columns in Zi.
 #' @param s A vector of initial values of the variance components, (s1, ...,sk, s_(k+1)).
+#' @param p Equal to 0 for ML, otherwise, ncol(X) for REML.
 #' @param max.iter The maximal number of iterations.
 #' @param epsilon Positive convergence tolerance.
 #'
@@ -162,9 +180,8 @@ list(method = method, dlogL = dlogL, niter = niter, coef = beta, se = sebeta, t
 #' @keywords internal
 #'
 #' @noRd
-varest <- function(xxz, zrz, zryj, yryj, n, d, s, max.iter = 50, epsilon = 1e-5)
+varest <- function(ZZres, zrz, zryj, yryj, n, d, s, p, max.iter = 50, epsilon = 1e-5)
 {
-  p <- nrow(xxz)
   k <- length(d)
 
   dl <- 100
@@ -177,38 +194,47 @@ varest <- function(xxz, zrz, zryj, yryj, n, d, s, max.iter = 50, epsilon = 1e-5)
 
     sr <- s[1:k]/s[k+1]
     M <- solve(sweep(zrz, 1, STATS = rep(sr, times = d), FUN = "*") + diag(sum(d)))
-    ZRZ <- zrz%*%M
-    ZR2Z <- ZRZ%*%M
     yRZ <- t(zryj)%*%M
-
+    
+    if (p > 0){
+    	M0 <- M
+    } else {
+    	M0 <- solve(sweep(ZZres, 1, STATS = rep(sr, times = d), FUN = "*") + diag(sum(d))) 	
+    }
+    ZVZ <- ZZres%*%M0
+    ZV2Z <- ZVZ%*%M0
+    
     mi <- 0
     for (i in 1:k){
       ik <- (mi+1):(mi+d[i])
-      dl[i] <- (sum((yRZ[ik])^2)/s[k+1]^2 - sum(diag(ZRZ[ik, ik, drop = FALSE]))/s[k+1])/2
+      dl[i] <- (sum((yRZ[ik])^2)/s[k+1]^2 - sum(diag(ZVZ[ik, ik, drop = FALSE]))/s[k+1])/2
 
       mj <- 0
       for (j in 1:i){
         ji <- (mj+1):(mj+d[j])
-        fs[i, j] <- sum((ZRZ[ji, ik])^2)/s[k+1]^2/2
+        fs[i, j] <- sum((ZVZ[ji, ik])^2)/s[k+1]^2/2
         fs[j, i] <- fs[i, j]
         mj <- mj + d[j]
       }
 
       j <- k+1
-      fs[i, j] <- sum(diag(ZR2Z[ik, ik, drop = FALSE]))/s[k+1]^2/2
+      fs[i, j] <- sum(diag(ZV2Z[ik, ik, drop = FALSE]))/s[k+1]^2/2
       fs[j, i] <- fs[i, j]
       mi <- mi + d[i]
     }
 
     i <- k+1
-    fs[i, i] <- (n - p - sum(d) + sum(t(M)*M))/s[k+1]^2/2
+    fs[i, i] <- (n - p - sum(d) + sum(t(M0)*M0))/s[k+1]^2/2
 
     yR2y <- yryj - sum(((t(M) + diag(sum(d)))%*%zryj)*(M%*%(rep(sr, times = d)*zryj)))
-    dl[i] <-  (yR2y/s[k+1]^2 - (n-p-sum(d)+sum(diag(M)))/s[k+1])/2
+    dl[i] <-  (yR2y/s[k+1]^2 - (n - p - sum(d) + sum(diag(M0)))/s[k+1])/2
 
     Minv <- ginv(fs)
     s <- s + Minv%*%dl
   }
 
-  list(s = c(s), dl = dl, iter = iter, Minv = Minv)
+qrM0 <- qr(M0)
+logdetM0 <- sum(log(abs(diag(qrM0$qr))))
+list(s = c(s), dl = dl, iter = iter, Minv = Minv, logdetM0 = logdetM0)
 }
+
